"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;

var _template = _interopRequireDefault(require("@babel/template"));

var _lexer = _interopRequireDefault(require("./lexer"));

var _parser = _interopRequireDefault(require("./parser"));

var _codeGen = _interopRequireDefault(require("./code-gen"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const buildHelper = (0, _template.default)(`
  function MERGE_QUERIES(firstQuery, secondQuery) {
    if (secondQuery === null) {
      return firstQuery;
    }
    const result = {};
    let firstQueryUpdated = false;
    let secondQueryUpdated = false;
    for (let key in firstQuery) {
      if (key in firstQuery) {
        if (!secondQuery[key]) {
          firstQueryUpdated = true;
          result[key] = firstQuery[key];
        } else if (firstQuery[key] !== secondQuery[key]) {
          result[key] = MERGE_QUERIES(firstQuery[key], secondQuery[key]);
          if (result[key] !== secondQuery[key]) {
            firstQueryUpdated = true;
            secondQueryUpdated = true;
          }
        } else {
          result[key] = firstQuery[key];
        }
      }
    }
    if (!firstQueryUpdated) {
      return secondQuery;
    }
    for (let key in secondQuery) {
      if ((key in secondQuery) && !result[key]) {
        secondQueryUpdated = true;
        result[key] = secondQuery[key];
      }
    }
    return secondQueryUpdated ? result : firstQuery;
  }
`);

function _default(babel) {
  const {
    types: t
  } = babel;

  function isBqlReference(node) {
    // TODO: do this better
    return t.isIdentifier(node, {
      name: 'bql'
    });
  }

  function isCallTo(node, name) {
    return t.isCallExpression(node) && t.isIdentifier(node.callee) && node.callee.name === name;
  } // check if a variable is every mutated


  const mutatedVisitor = {
    Identifier(path, state) {
      if (path.node.name === state.name) {
        const node = path.parentPath.node;
        const key = path.parentKey;
        const isMutation = t.isAssignmentExpression(node) && key === 'left' || t.isUpdateExpression(node);

        if (isMutation) {
          state.isMutated = true;
          path.stop();
        }
      }
    }

  }; // check if an expression is "safeish" to hoist

  const immutabilityVisitor = {
    ReferencedIdentifier(path, state) {
      if (path.node.name === 'JSON') {
        return;
      } // the only identifiers we'll treat as immutable are mergeBqlQueries and the queries being merged


      if (!state.mergeBqlQueries) {
        state.isImmutable = false;
        path.stop();
        return;
      }

      if (path.node.name === state.mergeBqlQueries.name) {
        return;
      }

      let parent = path.parentPath;

      while (!t.isStatement(parent.node) && !isCallTo(parent.node, state.mergeBqlQueries.name)) {
        if (t.isObjectExpression(parent.node)) {
          state.isImmutable = false;
          path.stop();
          return;
        }

        parent = parent.parentPath;
      }

      const s = {
        isMutated: false,
        name: path.node.name
      };
      state.filePath.traverse(mutatedVisitor, s);

      if (s.isMutated) {
        state.isImmutable = false;
        path.stop();
      }
    },

    enter(path, state) {
      if (path.isImmutable()) {
        return;
      }

      if (path.isIdentifier()) {
        return;
      }

      if (state.mergeBqlQueries && isCallTo(path.node, state.mergeBqlQueries.name)) {
        return;
      }

      if (t.isCallExpression(path.node) && t.isMemberExpression(path.node.callee) && t.isIdentifier(path.node.callee.object) && path.node.callee.object.name === 'JSON' && path.node.callee.computed === false && t.isIdentifier(path.node.callee.property) && path.node.callee.property.name === 'stringify') {
        return;
      } // see if the node is constant providing its children are constant


      const isConstant = t.isArrayExpression(path.node) || t.isBinaryExpression(path.node) || t.isBooleanLiteral(path.node) || t.isConditionalExpression(path.node) || t.isLogicalExpression(path.node) || t.isMemberExpression(path.node) || t.isNullLiteral(path.node) || t.isNumericLiteral(path.node) || t.isObjectExpression(path.node) || t.isObjectMethod(path.node) || t.isObjectProperty(path.node) || t.isParenthesizedExpression(path.node) || t.isSequenceExpression(path.node) || t.isSpreadElement(path.node) || t.isStringLiteral(path.node) || t.isTemplateLiteral(path.node) || false;

      if (isConstant) {
        return;
      }

      state.isImmutable = false;
      path.stop();
    }

  };

  function getMergeBqlQueries(file) {
    if (file._mergeBqlQueries) return file._mergeBqlQueries;
    file._mergeBqlQueries = file.scope.generateUidIdentifier('mergeBqlQueries');
    const helper = buildHelper({
      MERGE_QUERIES: file._mergeBqlQueries
    });
    helper.body._compact = true;
    helper._generated = true;
    file.path.unshiftContainer("body", helper);
    return file._mergeBqlQueries;
  }

  return {
    visitor: {
      TaggedTemplateExpression(path) {
        if (isBqlReference(path.node.tag)) {
          const tokens = (0, _lexer.default)(path.node.quasi, this.file);
          const ast = (0, _parser.default)(tokens, this.file);
          const transformed = (0, _codeGen.default)(t, ast, {
            file: this.file,
            getMergeBqlQueries: () => {
              return getMergeBqlQueries(this.file);
            }
          });
          path.replaceWith(transformed);
        }
      },

      ObjectExpression(path) {
        if (!path.node._isBqlOutput || path.node._hoisted) {
          return;
        }

        const state = {
          isImmutable: true,
          mergeBqlQueries: this.file._mergeBqlQueries,
          filePath: this.file.path
        };
        path.traverse(immutabilityVisitor, state);

        if (state.isImmutable) {
          path.hoist();
        } else {
          path.node._hoisted = true;
        }
      },

      CallExpression(path) {
        if (!path.node._isBqlOutput || path.node._hoisted) {
          return;
        }

        const state = {
          isImmutable: true,
          mergeBqlQueries: this.file._mergeBqlQueries,
          filePath: this.file.path
        };
        path.traverse(immutabilityVisitor, state);

        if (state.isImmutable) {
          path.hoist();
        } else {
          path.node._hoisted = true;
        }
      }

    }
  };
}